
"""Define a class to communicate with configuration database API."""

import json as _json
import datetime as _datetime
from urllib import parse as _parse
from urllib.request import Request as _Request, urlopen as _urlopen
from urllib.error import URLError as _URLError
import dateutil.parser
import numpy as _np

from .. import envars as _envars

from . import _templates


class ConfigDBClient:
    """Perform operation on configuration database."""

    _TIMEOUT_DEFAULT = 30.0
    _INVALID_CHARACTERS = '\\/:;,?!$'

    def __init__(self, url=None, config_type=None):
        """Class constructor.

        Parameters
        ----------
        url : str | None
            Configuration service host address. For default 'None' value
            the URL defined in siripy.envars is used.

        """
        self._url = url or _envars.SRVURL_CONFIGDB
        self._config_type = config_type

    @property
    def config_type(self):
        """Type of configuration."""
        return self._config_type

    @config_type.setter
    def config_type(self, name):
        if isinstance(name, str):
            self._config_type = name

    @property
    def url(self):
        """Server URL."""
        return self._url

    @property
    def connected(self):
        """Return connection state."""
        try:
            self.get_dbsize()
        except ConfigDBException as err:
            return not err.server_code == -2
        return True

    def get_dbsize(self):
        """Return estimated size of configuration database."""
        return self._make_request(stats=True)['size']

    def get_nrconfigs(self):
        """Return estimated size of configuration database."""
        return self._make_request(stats=True)['count']

    def get_config_types(self):
        """Get configuration types existing as database entries."""
        return self._make_request()

    @staticmethod
    def get_config_types_from_templates():
        """Return list of configuration types as defined in templates."""
        return list(_templates.get_config_types())

    def find_configs(self,
                     name=None,
                     begin=None,
                     end=None,
                     config_type=None,
                     discarded=False):
        """Find configurations matching search criteria.

        Parameters
        ----------
            discarded : True | False (default) | None
            If True, return only discarded configurations, if False, return
            only configurations in use. If None, return all configurations
            matching the other criteria.

        """
        config_type = self._process_config_type(config_type)
        # build search dictionary
        find_dict = dict(config_type=config_type)
        if name is not None:
            find_dict['name'] = name
        if begin is not None or end is not None:
            find_dict['created'] = {}
            if begin is not None:
                find_dict['created']['$gte'] = begin
            if end is not None:
                find_dict['created']['$lte'] = end

        return self._make_request(
            config_type=config_type, discarded=discarded, data=find_dict)

    def get_config_value(self, name, config_type=None, discarded=False):
        """Get value field of a given configuration."""
        config_type = self._process_config_type(config_type)
        return self._make_request(
            config_type=config_type, name=name, discarded=discarded)['value']

    def get_config_info(self, name, config_type=None, discarded=False):
        """Get information of a given configuration."""
        config_type = self._process_config_type(config_type)
        res = self.find_configs(
            name=name, config_type=config_type, discarded=discarded)
        if not res:
            raise ConfigDBException(
                {'code': 404, 'message': 'Configuration no found.'})
        return res[0]

    def rename_config(self, oldname, newname, config_type=None):
        """Rename configuration in database."""
        config_type = self._process_config_type(config_type)
        if not isinstance(newname, str):
            raise TypeError(
                'Config name must be str, not {}!'.format(type(newname)))
        if not self.check_valid_configname(newname):
            raise ValueError("There are invalid characters in config name!")

        return self._make_request(
            config_type=config_type, name=oldname, newname=newname,
            method='POST')

    def insert_config(self, name, value, config_type=None):
        """Insert configuration into database."""
        config_type = self._process_config_type(config_type)
        if not isinstance(name, str):
            raise TypeError(
                'Config name must be str, not {}!'.format(type(name)))
        if not self.check_valid_configname(name):
            raise ValueError("There are invalid characters in config name!")
        if not self.check_valid_value(value, config_type=config_type):
            raise TypeError('Incompatible configuration value!')

        self._make_request(
            config_type=config_type, name=name, method='POST', data=value)

    def delete_config(self, name, config_type=None):
        """Mark a valid configuration as discarded."""
        config_type = self._process_config_type(config_type)
        return self._make_request(
            config_type=config_type, name=name, method='DELETE')

    def retrieve_config(self, name, config_type=None):
        """Mark a discarded configuration as valid."""
        config_type = self._process_config_type(config_type)
        return self._make_request(
            config_type=config_type, name=name, discarded=True, method='POST')

    def get_value_from_template(self, config_type=None):
        """Return value of a configuration type."""
        config_type = self._process_config_type(config_type)
        return _templates.get_template(config_type)

    def check_valid_value(self, value, config_type=None):
        """Check whether values data corresponds to a configuration type."""
        config_type = self._process_config_type(config_type)
        return _templates.check_value(config_type, value)

    @classmethod
    def check_valid_configname(cls, name):
        "Check if `name` is a valid name for configurations."
        return not set(name) & set(cls._INVALID_CHARACTERS)

    @staticmethod
    def conv_timestamp_txt_2_flt(timestamp):
        """Convert timestamp format from text to float."""
        return dateutil.parser.parse(timestamp).timestamp()

    @staticmethod
    def conv_timestamp_flt_2_txt(timestamp):
        """Convert timestamp format from float to text."""
        return str(_datetime.datetime.fromtimestamp(timestamp))

    # --- private methods ---

    def _process_config_type(self, config_type):
        config_type = config_type or self._config_type
        if not config_type:
            raise ValueError(
                'You must define a `config_type` attribute or' +
                ' provide it in method call.')
        return config_type

    def _make_request(self, method='GET', data=None, **kwargs):
        try:
            return self._request(method, data, **kwargs)
        except ConfigDBException as err:
            if err.server_code == -2:
                self._rotate_server_url()
                return self._request(method, data, **kwargs)
            else:
                raise err

    def _request(self, method='GET', data=None, **kwargs):
        url = self._create_url(**kwargs)

        if data is None:
            request = _Request(url=url, method=method)
        else:
            request = _Request(
                url=url, method=method,
                headers={"Content-Type": "application/json"},
                data=_json.dumps(data, default=_jsonify_numpy).encode())

        try:
            url_conn = _urlopen(
                request, timeout=ConfigDBClient._TIMEOUT_DEFAULT)
            response = _json.loads(url_conn.read().decode("utf-8"))
        except _json.JSONDecodeError:
            response = {"code": -1, "message": "JSON decode error"}
        except _URLError as err:
            response = {'code': -2, 'message': str(err)}

        # print(response)
        if response['code'] != 200:
            raise ConfigDBException(response)
        return response['result']

    def _rotate_server_url(self):
        if self._url != _envars.SRVURL_CONFIGDB_2:
            self._url = _envars.SRVURL_CONFIGDB_2
        else:
            self._url = _envars.SRVURL_CONFIGDB

    def _create_url(self, config_type=None, name=None, discarded=False,
                    stats=False, newname=None):
        url = self.url
        if stats:
            return url + '/stats'
        url += '/configs'
        if newname:
            url += '/rename'
        if discarded:
            url += '/discarded'
        if config_type:
            url += '/' + config_type
        if name:
            url += '/' + name
        if newname:
            url += '/' + newname
        return _parse.quote(url, safe='/:')


class ConfigDBException(Exception):
    """Default exception raised for configDB server errors."""

    def __init__(self, response):
        """."""
        super().__init__('{code:d}: {message:s}.'.format(**response))
        self.server_code = response['code']
        self.server_message = response['message']


def _jsonify_numpy(obj):
    if isinstance(obj, _np.ndarray):
        return obj.tolist()
    raise TypeError('Object is not JSON serializable.')



"""BPM devices."""

import time as _time
# from threading import Event as _Flag

import numpy as _np

from .device import Device as _Device
from ..diagbeam.bpm.csdev import Const as _csbpm
from ..search import BPMSearch as _BPMSearch


class BPM(_Device):
    """BPM Device."""

    ACQSTATES_NOTOK = {
        _csbpm.AcqStates.Error, _csbpm.AcqStates.No_Memory,
        _csbpm.AcqStates.Too_Few_Samples,
        _csbpm.AcqStates.Too_Many_Samples, _csbpm.AcqStates.Acq_Overflow}
    ACQSTATES_STARTED = {
        _csbpm.AcqStates.Waiting, _csbpm.AcqStates.External_Trig,
        _csbpm.AcqStates.Data_Trig, _csbpm.AcqStates.Software_Trig,
        _csbpm.AcqStates.Acquiring}
    ACQSTATES_FINISHED = {_csbpm.AcqStates.Idle, _csbpm.AcqStates.Aborted}
    ACQSTATES_FINISHED |= ACQSTATES_NOTOK

    PROPERTIES_ACQ = (
        'INFOClkFreq-RB', 'INFOHarmonicNumber-RB', 'INFOTbTRate-RB',
        'INFOFOFBRate-RB', 'INFOMONITRate-RB', 'INFOFAcqRate-RB',
        'GEN_AArrayData', 'GEN_BArrayData', 'GEN_CArrayData', 'GEN_DArrayData',
        'GEN_XArrayData', 'GEN_YArrayData', 'GEN_SUMArrayData',
        'GEN_QArrayData',
        'ACQBPMMode-Sel', 'ACQBPMMode-Sts',
        'ACQChannel-Sel', 'ACQChannel-Sts',
        'ACQShots-SP', 'ACQShots-RB',
        'ACQUpdateTime-SP', 'ACQUpdateTime-RB',
        'ACQSamplesPre-SP', 'ACQSamplesPre-RB',
        'ACQSamplesPost-SP', 'ACQSamplesPost-RB',
        'ACQTriggerEvent-Sel', 'ACQTriggerEvent-Sts',
        'ACQStatus-Sts', 'ACQCount-Mon',
        'ACQTrigger-Sel', 'ACQTrigger-Sts',
        'ACQTriggerRep-Sel', 'ACQTriggerRep-Sts',
        'ACQDataTrigChan-Sel', 'ACQDataTrigChan-Sts',
        'ACQTriggerDataSel-SP', 'ACQTriggerDataSel-RB',
        'ACQTriggerDataThres-SP', 'ACQTriggerDataThres-RB',
        'ACQTriggerDataPol-Sel', 'ACQTriggerDataPol-Sts',
        'ACQTriggerDataHyst-SP', 'ACQTriggerDataHyst-RB',
        )

    PROPERTIES_DEFAULT = PROPERTIES_ACQ + (
        'asyn.ENBL', 'asyn.CNCT', 'SwMode-Sel', 'SwMode-Sts',
        'RFFEAtt-SP', 'RFFEAtt-RB',
        'SP_AArrayData', 'SP_BArrayData', 'SP_CArrayData', 'SP_DArrayData',
        'GEN_RawXArrayData', 'GEN_RawYArrayData', 'GEN_RawSUMArrayData',
        'GEN_RawQArrayData',
        'SPPosX-Mon', 'SPPosY-Mon', 'SPSum-Mon', 'SPPosQ-Mon',
        'SPAmplA-Mon', 'SPAmplB-Mon', 'SPAmplC-Mon', 'SPAmplD-Mon',
        'PosX-Mon', 'PosY-Mon', 'Sum-Mon', 'PosQ-Mon',
        'AmplA-Mon', 'AmplB-Mon', 'AmplC-Mon', 'AmplD-Mon',
        'GEN_PolyXArrayCoeff-SP', 'GEN_PolyXArrayCoeff-RB',
        'GEN_PolyYArrayCoeff-SP', 'GEN_PolyYArrayCoeff-RB',
        'GEN_PolySUMArrayCoeff-SP', 'GEN_PolySUMArrayCoeff-RB',
        'GEN_PolyQArrayCoeff-SP', 'GEN_PolyQArrayCoeff-RB',
        'SwDirGainA-SP', 'SwDirGainB-SP', 'SwDirGainC-SP', 'SwDirGainD-SP',
        'SwDirGainA-RB', 'SwDirGainB-RB', 'SwDirGainC-RB', 'SwDirGainD-RB',
        'SwInvGainA-SP', 'SwInvGainB-SP', 'SwInvGainC-SP', 'SwInvGainD-SP',
        'SwInvGainA-RB', 'SwInvGainB-RB', 'SwInvGainC-RB', 'SwInvGainD-RB',
        'PosKx-SP', 'PosKx-RB',
        'PosKy-RB', 'PosKy-SP',
        'PosKsum-SP', 'PosKsum-RB',
        'PosKq-SP', 'PosKq-RB',
        'PosXOffset-SP', 'PosXOffset-RB',
        'PosYOffset-SP', 'PosYOffset-RB',
        'PosSumOffset-SP', 'PosSumOffset-RB',
        'PosQOffset-SP', 'PosQOffset-RB',
        'FOFBPhaseSyncEn-Sel', 'FOFBPhaseSyncEn-Sts', 'SwDivClk-RB',
        'TbTPhaseSyncEn-Sel', 'TbTPhaseSyncEn-Sts',
        'FAcqPhaseSyncEn-Sel', 'FAcqPhaseSyncEn-Sts',
        'MonitPhaseSyncEn-Sel', 'MonitPhaseSyncEn-Sts',
        'TbTDataMaskEn-Sel', 'TbTDataMaskEn-Sts',
        'TbTDataMaskSamplesBeg-SP', 'TbTDataMaskSamplesBeg-RB',
        'TbTDataMaskSamplesEnd-SP', 'TbTDataMaskSamplesEnd-RB',
        'XYPosCal-Sel', 'XYPosCal-Sts',
        'SUMPosCal-Sel', 'SUMPosCal-Sts',
        'QPosCal-Sel', 'QPosCal-Sts',
        )

    CONV_NM2UM = 1e-3  # [nm] --> [um]

    def __init__(
            self, devname, props2init='all', auto_monitor_mon=True,
            ispost_mortem=False):
        """."""
        # call base class constructor
        self._ispost_mortem = ispost_mortem

        if props2init == 'all' and _BPMSearch.is_photon_bpm(devname):
            props2init = set(BPM.PROPERTIES_DEFAULT)
            props2init -= {'RFFEAtt-SP', 'RFFEAtt-RB'}
            props2init = list(props2init)
        elif isinstance(props2init, str) and props2init.startswith('acq'):
            props2init = list(self.PROPERTIES_ACQ)

        super().__init__(
            devname, props2init=props2init, auto_monitor_mon=auto_monitor_mon)
        self.csdata = _csbpm

    def __str__(self):
        """."""
        stg = '################### Summary Status ###################\n'
        stg += 'asyn:\n'
        stg += f'    Enabled: {_csbpm.EnblTyp._fields[self.asyn_state]:s}\n'
        stg += '    Connected: '
        stg += '    Switching Mode: '
        stg += f'{_csbpm.SwModes._fields[self.switching_mode]:s}\n'
        stg += f'{_csbpm.ConnTyp._fields[self.asyn_connected]:s}\n'
        stg += '\nAcquisition Parameters:\n'
        stg += f'    - Status: {_csbpm.AcqStates._fields[self.acq_status]:s}\n'
        stg += f'    - Mode: {_csbpm.OpModes._fields[self.acq_mode]:s}\n'
        stg += f'    - Channel: {_csbpm.AcqChan._fields[self.acq_channel]:s}\n'
        stg += f'    - Nr Shots: {self.acq_nrshots:d}\n'
        stg += f'    - Update Time: {self.acq_update_time:.1f} ms\n'
        stg += f'    - Repeat: {_csbpm.AcqRepeat._fields[self.acq_repeat]:s}\n'
        stg += '    - Trigger Type: '
        stg += f'{_csbpm.AcqTrigTyp._fields[self.acq_trigger]:s}\n'
        if self.acq_trigger == _csbpm.AcqTrigTyp.Data:
            stg += '        - Channel: '
            stg += f'{_csbpm.AcqChan._fields[self.acq_trig_datachan]:s}\n'
            stg += '        - Source: '
            stg += f'{_csbpm.AcqDataTyp._fields[self.acq_trig_datasel]:s}\n'
            stg += '        - Polarity: '
            stg += f'{_csbpm.Polarity._fields[self.acq_trig_datapol]:s}\n'
            stg += f'        - Threshold: {self.acq_trig_datathres:.1f}\n'
            stg += f'        - Hysteresis: {self.acq_trig_datahyst:d}\n'
        stg += '\n'
        return stg

    @property
    def is_ok(self):
        """."""
        okay = self.acq_status not in self.ACQSTATES_NOTOK
        okay &= self.asyn_connected == _csbpm.ConnTyp.Connected
        okay &= self.asyn_state == _csbpm.EnblTyp.Enable
        return okay

    @property
    def rffe_att(self):
        """."""
        if 'RFFEAtt-RB' not in self._pvs:
            return None
        return self['RFFEAtt-RB']

    @rffe_att.setter
    def rffe_att(self, val):
        """."""
        if 'RFFEAtt-SP' in self._pvs:
            self['RFFEAtt-SP'] = val

    @property
    def asyn_state(self):
        """."""
        return self['asyn.ENBL']

    @asyn_state.setter
    def asyn_state(self, boo):
        """."""
        val = _csbpm.EnblTyp.Enable if boo else _csbpm.EnblTyp.Disable
        self['asyn.ENBL'] = val

    @property
    def asyn_connected(self):
        """."""
        return self['asyn.CNCT']

    @property
    def switching_mode(self):
        """."""
        return self['SwMode-Sts']

    @switching_mode.setter
    def switching_mode(self, val):
        """."""
        self['SwMode-Sel'] = val

    @property
    def switching_mode_str(self):
        """."""
        return _csbpm.SwModes._fields[self.switching_mode]

    @property
    def switching_rate(self):
        """Switching rate manifested at BPM readings.

        NOTE: Actually this property maps the rate of the effect of
        switching on the orbit, which is twice the switching rate,
        because the switching has two states.

        """
        return self['SwDivClk-RB'] * 2

    @property
    def switching_period(self):
        """Switching period manifested at BPM readings."""
        return self.switching_rate / self.adcfreq

    @property
    def gain_direct_a(self):
        """."""
        return self['SwDirGainA-RB']

    @gain_direct_a.setter
    def gain_direct_a(self, val):
        self['SwDirGainA-SP'] = val

    @property
    def gain_direct_b(self):
        """."""
        return self['SwDirGainB-RB']

    @gain_direct_b.setter
    def gain_direct_b(self, val):
        self['SwDirGainB-SP'] = val

    @property
    def gain_direct_c(self):
        """."""
        return self['SwDirGainC-RB']

    @gain_direct_c.setter
    def gain_direct_c(self, val):
        self['SwDirGainC-SP'] = val

    @property
    def gain_direct_d(self):
        """."""
        return self['SwDirGainD-RB']

    @gain_direct_d.setter
    def gain_direct_d(self, val):
        self['SwDirGainD-SP'] = val

    @property
    def gain_inverse_a(self):
        """."""
        return self['SwInvGainA-RB']

    @gain_inverse_a.setter
    def gain_inverse_a(self, val):
        self['SwInvGainA-SP'] = val

    @property
    def gain_inverse_b(self):
        """."""
        return self['SwInvGainB-RB']

    @gain_inverse_b.setter
    def gain_inverse_b(self, val):
        self['SwInvGainB-SP'] = val

    @property
    def gain_inverse_c(self):
        """."""
        return self['SwInvGainC-RB']

    @gain_inverse_c.setter
    def gain_inverse_c(self, val):
        self['SwInvGainC-SP'] = val

    @property
    def gain_inverse_d(self):
        """."""
        return self['SwInvGainD-RB']

    @gain_inverse_d.setter
    def gain_inverse_d(self, val):
        self['SwInvGainD-SP'] = val

    @property
    def harmonic_number(self):
        """."""
        return self['INFOHarmonicNumber-RB']

    @property
    def adcfreq(self):
        """."""
        return self['INFOClkFreq-RB']

    @property
    def tbt_rate(self):
        """Divisor or TbT in relation to ADC."""
        return self['INFOTbTRate-RB']

    @property
    def tbt_period(self):
        """."""
        return self.tbt_rate / self.adcfreq

    @property
    def tbt_sync_enbl(self):
        """."""
        return self['TbTPhaseSyncEn-Sts']

    @tbt_sync_enbl.setter
    def tbt_sync_enbl(self, val):
        """."""
        self['TbTPhaseSyncEn-Sel'] = val

    @property
    def tbt_mask_enbl(self):
        """."""
        return self['TbTDataMaskEn-Sts']

    @tbt_mask_enbl.setter
    def tbt_mask_enbl(self, val):
        """."""
        self['TbTDataMaskEn-Sel'] = val

    @property
    def tbt_mask_beg(self):
        """."""
        return self['TbTDataMaskSamplesBeg-RB']

    @tbt_mask_beg.setter
    def tbt_mask_beg(self, val):
        """."""
        self['TbTDataMaskSamplesBeg-SP'] = val

    @property
    def tbt_mask_end(self):
        """."""
        return self['TbTDataMaskSamplesEnd-RB']

    @tbt_mask_end.setter
    def tbt_mask_end(self, val):
        """."""
        self['TbTDataMaskSamplesEnd-SP'] = val

    @property
    def fofb_sync_enbl(self):
        """."""
        return self['FOFBPhaseSyncEn-Sts']

    @fofb_sync_enbl.setter
    def fofb_sync_enbl(self, val):
        """."""
        self['FOFBPhaseSyncEn-Sel'] = val

    @property
    def fofb_rate(self):
        """Divisor or FOFB in relation to ADC."""
        return self['INFOFOFBRate-RB']

    @property
    def fofb_period(self):
        """."""
        return self.fofb_rate / self.adcfreq

    @property
    def facq_rate(self):
        """Divisor or FAcq in relation to ADC."""
        return self['INFOFAcqRate-RB']

    @property
    def facq_period(self):
        """."""
        return self.facq_rate / self.adcfreq

    @property
    def facq_sync_enbl(self):
        """."""
        return self['FAcqPhaseSyncEn-Sts']

    @facq_sync_enbl.setter
    def facq_sync_enbl(self, val):
        """."""
        self['FAcqPhaseSyncEn-Sel'] = val

    @property
    def monit_rate(self):
        """Divisor or Monit in relation to ADC."""
        return self['INFOMONITRate-RB']

    @property
    def monit_period(self):
        """."""
        return self.monit_rate / self.adcfreq

    @property
    def monit_sync_enbl(self):
        """."""
        return self['MonitPhaseSyncEn-Sts']

    @monit_sync_enbl.setter
    def monit_sync_enbl(self, val):
        """."""
        self['MonitPhaseSyncEn-Sel'] = val

    @property
    def posx_gain(self):
        """."""
        return self['PosKx-RB'] * self.CONV_NM2UM

    @posx_gain.setter
    def posx_gain(self, value):
        self['PosKx-SP'] = float(value) / self.CONV_NM2UM

    @property
    def posy_gain(self):
        """."""
        return self['PosKy-RB'] * self.CONV_NM2UM

    @posy_gain.setter
    def posy_gain(self, value):
        self['PosKy-SP'] = float(value) / self.CONV_NM2UM

    @property
    def possum_gain(self):
        """."""
        return self['PosKsum-RB']

    @possum_gain.setter
    def possum_gain(self, value):
        self['PosKsum-SP'] = float(value)

    @property
    def posq_gain(self):
        """."""
        return self['PosKq-RB']

    @posq_gain.setter
    def posq_gain(self, value):
        self['PosKq-SP'] = float(value)

    @property
    def posx_offset(self):
        """."""
        return self['PosXOffset-RB'] * self.CONV_NM2UM

    @posx_offset.setter
    def posx_offset(self, value):
        self['PosXOffset-SP'] = float(value) / self.CONV_NM2UM

    @property
    def posy_offset(self):
        """."""
        return self['PosYOffset-RB'] * self.CONV_NM2UM

    @posy_offset.setter
    def posy_offset(self, value):
        self['PosYOffset-SP'] = float(value) / self.CONV_NM2UM

    @property
    def possum_offset(self):
        """."""
        return self['PosSumOffset-RB']

    @possum_offset.setter
    def possum_offset(self, value):
        self['PosSumOffset-SP'] = float(value)

    @property
    def posq_offset(self):
        """."""
        return self['PosQOffset-RB']

    @posq_offset.setter
    def posq_offset(self, value):
        self['PosQOffset-SP'] = float(value)

    @property
    def posx(self):
        """."""
        return self['PosX-Mon'] * self.CONV_NM2UM

    @property
    def posy(self):
        """."""
        return self['PosY-Mon'] * self.CONV_NM2UM

    @property
    def possum(self):
        """."""
        return self['Sum-Mon']

    @property
    def posq(self):
        """."""
        return self['PosQ-Mon']

    @property
    def ampla(self):
        """."""
        return self['AmplA-Mon']

    @property
    def amplb(self):
        """."""
        return self['AmplB-Mon']

    @property
    def amplc(self):
        """."""
        return self['AmplC-Mon']

    @property
    def ampld(self):
        """."""
        return self['AmplD-Mon']

    @property
    def mt_posx(self):
        """."""
        return self['GEN_XArrayData'] * self.CONV_NM2UM

    @property
    def mt_posy(self):
        """."""
        return self['GEN_YArrayData'] * self.CONV_NM2UM

    @property
    def mt_possum(self):
        """."""
        return self['GEN_SUMArrayData']

    @property
    def mt_posq(self):
        """."""
        return self['GEN_QArrayData']

    @property
    def mt_ampla(self):
        """."""
        return self['GEN_AArrayData']

    @property
    def mt_amplb(self):
        """."""
        return self['GEN_BArrayData']

    @property
    def mt_amplc(self):
        """."""
        return self['GEN_CArrayData']

    @property
    def mt_ampld(self):
        """."""
        return self['GEN_DArrayData']

    @property
    def mt_polyx(self):
        """."""
        return self['GEN_PolyXArrayCoeff-RB']

    @mt_polyx.setter
    def mt_polyx(self, value):
        """."""
        self['GEN_PolyXArrayCoeff-SP'] = _np.array(value)

    @property
    def mt_polyy(self):
        """."""
        return self['GEN_PolyYArrayCoeff-RB']

    @mt_polyy.setter
    def mt_polyy(self, value):
        """."""
        self['GEN_PolyYArrayCoeff-SP'] = _np.array(value)

    @property
    def mt_polysum(self):
        """."""
        return self['GEN_PolySUMArrayCoeff-RB']

    @mt_polysum.setter
    def mt_polysum(self, value):
        """."""
        self['GEN_PolySUMArrayCoeff-SP'] = _np.array(value)

    @property
    def mt_polyq(self):
        """."""
        return self['GEN_PolyQArrayCoeff-RB']

    @mt_polyq.setter
    def mt_polyq(self, value):
        """."""
        self['GEN_PolyQArrayCoeff-SP'] = _np.array(value)

    @property
    def mt_polyxy_enbl(self):
        """."""
        return self['XYPosCal-Sts']

    @mt_polyxy_enbl.setter
    def mt_polyxy_enbl(self, val):
        """."""
        self['XYPosCal-Sel'] = val

    @property
    def mt_polysum_enbl(self):
        """."""
        return self['SUMPosCal-Sts']

    @mt_polysum_enbl.setter
    def mt_polysum_enbl(self, val):
        """."""
        self['SUMPosCal-Sel'] = val

    @property
    def mt_polyq_enbl(self):
        """."""
        return self['QPosCal-Sts']

    @mt_polyq_enbl.setter
    def mt_polyq_enbl(self, val):
        """."""
        self['QPosCal-Sel'] = val

    @property
    def mtraw_posx(self):
        """Multi turn raw X array data."""
        return self['GEN_RawXArrayData'] * self.CONV_NM2UM

    @property
    def mtraw_posy(self):
        """Multi turn raw Y array data."""
        return self['GEN_RawYArrayData'] * self.CONV_NM2UM

    @property
    def mtraw_possum(self):
        """Multi turn raw sum array data."""
        return self['GEN_RawSUMArrayData']

    @property
    def mtraw_posq(self):
        """Multi turn raw Q array data."""
        return self['GEN_RawQArrayData']

    @property
    def sp_posx(self):
        """."""
        return self['SPPosX-Mon'] * self.CONV_NM2UM

    @property
    def sp_posy(self):
        """."""
        return self['SPPosY-Mon'] * self.CONV_NM2UM

    @property
    def sp_possum(self):
        """."""
        return self['SPSum-Mon']

    @property
    def sp_posq(self):
        """."""
        return self['SPPosQ-Mon']

    @property
    def sp_ampla(self):
        """."""
        return self['SPAmplA-Mon']

    @property
    def sp_amplb(self):
        """."""
        return self['SPAmplB-Mon']

    @property
    def sp_amplc(self):
        """."""
        return self['SPAmplC-Mon']

    @property
    def sp_ampld(self):
        """."""
        return self['SPAmplD-Mon']

    @property
    def sp_arraya(self):
        """."""
        return self['SP_AArrayData']

    @property
    def sp_arrayb(self):
        """."""
        return self['SP_BArrayData']

    @property
    def sp_arrayc(self):
        """."""
        return self['SP_CArrayData']

    @property
    def sp_arrayd(self):
        """."""
        return self['SP_DArrayData']

    @property
    def acq_mode(self):
        """."""
        return self['ACQBPMMode-Sts']

    @acq_mode.setter
    def acq_mode(self, mode):
        """."""
        self['ACQBPMMode-Sel'] = mode

    @property
    def acq_ctrl(self):
        """."""
        return self['ACQTriggerEvent-Sts']

    @acq_ctrl.setter
    def acq_ctrl(self, val):
        """."""
        self['ACQTriggerEvent-Sel'] = val

    @property
    def acq_status(self):
        """."""
        return self['ACQStatus-Sts']

    @property
    def acq_count(self):
        """Counter of number of acquisitions so far."""
        return self['ACQCount-Mon']

    @property
    def acq_channel(self):
        """."""
        return self['ACQChannel-Sts']

    @acq_channel.setter
    def acq_channel(self, val):
        """."""
        self['ACQChannel-Sel'] = val

    @property
    def acq_channel_str(self):
        """."""
        return _csbpm.AcqChan._fields[self.acq_channel]

    @property
    def acq_trigger(self):
        """."""
        return self['ACQTrigger-Sts']

    @acq_trigger.setter
    def acq_trigger(self, val):
        """."""
        self['ACQTrigger-Sel'] = val

    @property
    def acq_repeat(self):
        """."""
        return self['ACQTriggerRep-Sts']

    @acq_repeat.setter
    def acq_repeat(self, val):
        """."""
        self['ACQTriggerRep-Sel'] = val

    @property
    def acq_update_time(self):
        """BPMs update time in [s]."""
        return self['ACQUpdateTime-RB']

    @acq_update_time.setter
    def acq_update_time(self, val):
        """BPMs update time in [s]."""
        self['ACQUpdateTime-SP'] = val

    @property
    def acq_trig_datachan(self):
        """."""
        return self['ACQDataTrigChan-Sts']

    @acq_trig_datachan.setter
    def acq_trig_datachan(self, val):
        """."""
        self['ACQDataTrigChan-Sel'] = val

    @property
    def acq_trig_datasel(self):
        """."""
        return self['ACQTriggerDataSel-RB']

    @acq_trig_datasel.setter
    def acq_trig_datasel(self, val):
        """."""
        self['ACQTriggerDataSel-SP'] = val

    @property
    def acq_trig_datathres(self):
        """."""
        return self['ACQTriggerDataThres-RB']

    @acq_trig_datathres.setter
    def acq_trig_datathres(self, val):
        """."""
        self['ACQTriggerDataThres-SP'] = val

    @property
    def acq_trig_datahyst(self):
        """."""
        return self['ACQTriggerDataHyst-RB']

    @acq_trig_datahyst.setter
    def acq_trig_datahyst(self, val):
        """."""
        self['ACQTriggerDataHyst-SP'] = val

    @property
    def acq_trig_datapol(self):
        """."""
        return self['ACQTriggerDataPol-RB']

    @acq_trig_datapol.setter
    def acq_trig_datapol(self, val):
        """."""
        self['ACQTriggerDataPol-SP'] = val

    @property
    def acq_nrsamples_post(self):
        """."""
        return self['ACQSamplesPost-RB']

    @acq_nrsamples_post.setter
    def acq_nrsamples_post(self, val):
        """."""
        self['ACQSamplesPost-SP'] = val

    @property
    def acq_nrsamples_pre(self):
        """."""
        return self['ACQSamplesPre-RB']

    @acq_nrsamples_pre.setter
    def acq_nrsamples_pre(self, val):
        """."""
        self['ACQSamplesPre-SP'] = val

    @property
    def acq_nrshots(self):
        """."""
        return self['ACQShots-RB']

    @acq_nrshots.setter
    def acq_nrshots(self, val):
        """."""
        self['ACQShots-SP'] = val

    def wait_acq_finish(self, timeout=10):
        """Wait Acquisition to finish."""
        return self._wait(
            'ACQStatus-Sts', self.ACQSTATES_FINISHED, timeout=timeout,
            comp=lambda x, y: x in y)

    def wait_acq_start(self, timeout=10):
        """Wait Acquisition to start."""
        return self._wait(
            'ACQStatus-Sts', self.ACQSTATES_STARTED, timeout=timeout,
            comp=lambda x, y: x in y)

    def cmd_acq_start(self):
        """Command Start Acquisition."""
        self.acq_ctrl = _csbpm.AcqEvents.Start
        return self._wait('ACQTriggerEvent-Sts', _csbpm.AcqEvents.Start)

    def cmd_acq_stop(self):
        """Command Stop Acquisition."""
        self.acq_ctrl = _csbpm.AcqEvents.Stop
        return self._wait('ACQTriggerEvent-Sts', _csbpm.AcqEvents.Stop)

    def cmd_acq_abort(self):
        """Command Abort Acquisition."""
        self.acq_ctrl = _csbpm.AcqEvents.Abort
        return self._wait('ACQTriggerEvent-Sts', _csbpm.AcqEvents.Abort)

    def cmd_turn_on_switching(self):
        """Command Turn on Switching."""
        self.switching_mode = _csbpm.SwModes.switching
        return self._wait('SwMode-Sts', _csbpm.SwModes.switching)

    def cmd_turn_off_switching(self):
        """Command Turn off Switching."""
        self.switching_mode = _csbpm.SwModes.direct
        return self._wait('SwMode-Sts', _csbpm.SwModes.direct)

    def cmd_sync_tbt(self):
        """Synchronize TbT acquisitions with Timing System."""
        self.tbt_sync_enbl = 1
        _time.sleep(0.1)
        self.tbt_sync_enbl = 0
        return self._wait('TbTPhaseSyncEn-Sts', 0)

    def cmd_sync_fofb(self):
        """Synchronize FOFB acquisitions with Timing System."""
        self.fofb_sync_enbl = 1
        _time.sleep(0.1)
        self.fofb_sync_enbl = 0
        return self._wait('FOFBPhaseSyncEn-Sts', 0)

    def cmd_sync_facq(self):
        """Synchronize FAcq acquisitions with Timing System."""
        self.facq_sync_enbl = 1
        _time.sleep(0.1)
        self.facq_sync_enbl = 0
        return self._wait('FAcqPhaseSyncEn-Sts', 0)

    def cmd_sync_monit(self):
        """Synchronize Monit acquisitions with Timing System."""
        self.monit_sync_enbl = 1
        _time.sleep(0.1)
        self.monit_sync_enbl = 0
        return self._wait('MonitPhaseSyncEn-Sts', 0)

    def get_sampling_frequency(
            self, rf_freq: float, acq_rate='') -> float:
        """Return the sampling frequency of the acquisition.

        Args:
            rf_freq (float): RF frequency.
            acq_rate (str, optional): acquisition rate. Defaults to ''.
            If empty string, it gets the configured acq. rate on BPMs

        Returns:
            float: acquisition frequency.

        """
        acq_rate = self.acq_channel_str if not acq_rate else acq_rate
        fadc = rf_freq / self.harmonic_number * self.tbt_rate
        if acq_rate.lower().startswith('tbt'):
            return fadc / self.tbt_rate
        elif acq_rate.lower().startswith('fofb'):
            return fadc / self.fofb_rate
        elif acq_rate.lower().startswith('facq'):
            return fadc / self.facq_rate
        return fadc / self.monit_rate

    def get_switching_frequency(self, rf_freq: float) -> float:
        """Return the switching frequency.

        Args:
            rf_freq (float): RF frequency.

        Returns:
            float: switching frequency.

        """
        fadc = rf_freq / self.harmonic_number * self.tbt_rate
        return fadc / self.switching_rate

    def get_propname(self, prop):
        """Get appropriate property name in case of triggered acquisitions."""
        if not self._ispost_mortem:
            return prop
        if prop.startswith('GEN'):
            return prop.replace('GEN', 'PM')
        elif prop.startswith('ACQ'):
            return prop.replace('ACQ', 'ACQ_PM')
        return prop

    def _get_pvname(self, propty):
        propty = self.get_propname(propty)
        return super()._get_pvname(propty)



"""Main module of Machine Shift Application."""
import time as _time
from datetime import datetime as _datetime

from ..callbacks import Callback as _Callback

from .csdev import Const as _Const, ETypes as _ETypes, \
    get_machshift_propty_database as _get_database
from .macschedule import MacScheduleData as _MacSched
from .utils import HandleMachShiftFile as _HandleMachShiftFile


class App(_Callback):
    """Main application for handling machine shift."""

    def __init__(self):
        """Class constructor."""
        super().__init__()
        self._pvs_database = _get_database()

        self._progmd_users = _MacSched.is_user_shift_programmed(
            datetime=_datetime.now())

        self.ms_handler = _HandleMachShiftFile()
        try:
            self._mode = self.ms_handler.get_machshift()
        except Exception:
            self._mode = _Const.MachShift.Commissioning

    def init_database(self):
        """Set initial PV values."""
        self.run_callbacks('Mode-Sel', self._mode)
        self.run_callbacks('Mode-Sts', self._mode)
        self.run_callbacks('ProgmdUsersShift-Mon', self._progmd_users)

    @property
    def pvs_database(self):
        """Return pvs_database."""
        return self._pvs_database

    def process(self, interval):
        """Sleep."""
        _time.sleep(interval)

    def read(self, reason):
        """Read from IOC database."""
        value = None
        if 'ProgmdUsersShift' in reason:
            new_progmd_users = _MacSched.is_user_shift_programmed(
                datetime=_datetime.now())
            if new_progmd_users != self._progmd_users:
                self._progmd_users = new_progmd_users
                value = new_progmd_users
                self.run_callbacks('ProgmdUsersShift-Mon', self._progmd_users)
        return value

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        status = False
        if reason == 'Mode-Sel':
            if 0 <= value < len(_ETypes.MACHSHIFT):
                self._mode = value
                self.run_callbacks('Mode-Sts', value)
                self.ms_handler.set_machshift(value)
                status = True
        return status



"""ID feedforward App."""

import os as _os
import logging as _log
import time as _time
from importlib.util import find_spec as _find_spec

import epics as _epics
import numpy as _np

from ..util import update_bit as _updt_bit
from ..callbacks import Callback as _Callback
from ..clientconfigdb import ConfigDBException as _ConfigDBException
from ..devices import IDFF as _IDFF
from ..pwrsupply.pssofb import PSConnSOFB as _PSConnSOFB
from ..pwrsupply.pssofb import PSNamesSOFB as _PSNamesSOFB

from .csdev import IDFFConst as _Const, ETypes as _ETypes

if _find_spec('PRUserial485') is not None:
    from PRUserial485 import EthBridgeClient as _EthBridgeClient


class App(_Callback):
    """Main application for handling IDFF."""

    DEF_PS_TIMEOUT = 5  # [s]

    def __init__(self, idname):
        """Class constructor."""
        super().__init__()
        self._const = _Const(idname)
        self._pvs_prefix = self._const.idffname
        self._pvs_database = self._const.get_propty_database()

        self._loop_state = _Const.DsblEnbl.Dsbl
        self._loop_freq = _Const.DEFAULT_LOOP_FREQ
        self._control_qs = _Const.DsblEnbl.Dsbl
        self._polarization = 'none'
        self._config_name = ''
        self.read_autosave_file()

        self._idff = _IDFF(idname)

        self._pssofb_isused = self._pvs_database['SOFBMode-Sts']['value']
        self._pssofb, self._bsmp_devs = self._pssofb_init(idname)

        # load idff in configdb
        self._load_config(self._config_name)

        # pvs to write methods
        self.map_pv2write = {
            'LoopState-Sel': self.set_loop_state,
            'LoopFreq-SP': self.set_loop_freq,
            'ControlQS-Sel': self.set_control_qs,
            'ConfigName-SP': self.set_config_name,
            'SOFBMode-Sel': self.set_sofb_mode,
            'CorrConfig-Cmd': self.cmd_corrconfig,
        }

        self._quit = False
        self._thread_ff = _epics.ca.CAThread(
            target=self._do_ff, daemon=True)
        self._thread_ff.start()

    def init_database(self):
        """Set initial PV values."""
        pvn2vals = {
            'LoopState-Sel': self._loop_state,
            'LoopState-Sts': self._loop_state,
            'LoopFreq-SP': self._loop_freq,
            'LoopFreq-RB': self._loop_freq,
            'ConfigName-SP': self._config_name,
            'ConfigName-RB': self._config_name,
            'Polarization-Mon': self._polarization,
            'SOFBMode-Sel': self._pssofb_isused,
            'SOFBMode-Sts': self._pssofb_isused,
            'CorrConfig-Cmd': 0,
            'CorrStatus-Mon': 0b1111,
        }
        if self._const.has_qscorrs:
            pvn2vals.update({
                'ControlQS-Sel': self._control_qs,
                'ControlQS-Sts': self._control_qs,
                })
        for pvn, val in pvn2vals.items():
            self.run_callbacks(pvn, val)
        self._update_log('Started.')

    @property
    def pvs_prefix(self):
        """Return pvs_prefix."""
        return self._pvs_prefix

    @property
    def pvs_database(self):
        """Return pvs_database."""
        return self._pvs_database

    def process(self, interval):
        """Sleep."""
        # check correctors state periodically
        _t0 = _time.time()
        self._update_corr_status()
        dtime = _time.time() - _t0
        sleep_time = interval - dtime
        # sleep
        if sleep_time > 0:
            _time.sleep(sleep_time)
        else:
            _log.debug('process took {0:f}ms.'.format((dtime)*1000))

    def read(self, reason):
        """Read from IOC database."""
        value = None
        return value

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        _log.info('Write received for: %s --> %s', reason, str(value))
        if reason in self.map_pv2write.keys():
            status = self.map_pv2write[reason](value)
            _log.info('%s Write for: %s --> %s',
                      str(status).upper(), reason, str(value))
            return status
        _log.warning('PV %s does not have a set function.', reason)
        return False

    def set_loop_state(self, value):
        """Set loop state."""
        if not 0 <= value < len(_ETypes.OPEN_CLOSED):
            return False
        self._loop_state = value
        act = ('En' if value else 'Dis')
        self._update_log(f'Feedforward loop {act}abled.')
        self.run_callbacks('LoopState-Sts', value)
        return True

    def set_loop_freq(self, value):
        """Set loop frequency."""
        if not 1e-3 <= value < 60:
            return False
        self._loop_freq = value
        self._update_log(f'Loop frequency updated to {value:.2f}Hz.')
        self.run_callbacks('LoopFreq-RB', value)
        return True

    def set_control_qs(self, value):
        """Set whether to include QS or not in feedforward."""
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            return False
        self._control_qs = value
        act = ('En' if value else 'Dis')
        self._update_log(f'{act}abled QS control.')
        self.run_callbacks('ControlQS-Sts', value)
        return True

    def set_config_name(self, value):
        """Set configuration name."""
        if self._loop_state == self._const.LoopState.Closed:
            self._update_log('ERR:Open loop before changing configuration.')
            return False

        if not self._load_config(value):
            return False

        self._config_name = value
        self.update_autosave_file()
        self.run_callbacks('ConfigName-RB', value)
        return True

    def set_sofb_mode(self, value):
        """Set whether to use SOFBMode."""
        if not 0 <= value < len(_ETypes.DSBL_ENBL):
            return False

        if self._loop_state == self._const.LoopState.Closed:
            self._update_log('ERR:Open loop before changing configuration.')
            return False

        if not self._idff_prepare_corrs_state(value):
            self._update_log(
                ('ERR:Could not configure IDFF correctors '
                 'when changing SOFBMode.'))
            return False

        self._pssofb_isused = bool(value)
        status = 'enabled' if self._pssofb_isused else 'disabled'
        self._update_log(f'SOFBMode {status}.')
        self.run_callbacks('SOFBMode-Sts', value)

        return True

    def cmd_corrconfig(self, _):
        """Command to reconfigure power supplies to desired state."""
        if self._loop_state == self._const.LoopState.Closed:
            self._update_log('ERR:Open loop before configure correctors.')
            return False
        if self._pssofb_isused:
            self._update_log('ERR:Turn off PSSOFB mode before configure.')
            return False

        corrdevs = self._idff.chdevs + self._idff.cvdevs + self._idff.qsdevs
        for dev in corrdevs:
            # turn on
            if not dev.cmd_turn_on():
                return False
            # opmode slowref
            if not dev.cmd_slowref():
                return False

        return True

    def _load_config(self, config_name):
        try:
            self._idff.load_config(config_name)
            self._update_log(f'Updated configuration: {config_name}.')
        except (ValueError, _ConfigDBException) as err:
            self._update_log('ERR:'+str(err))
            return False
        return True

    @property
    def quit(self):
        """Quit and shutdown threads."""
        return self._quit

    @quit.setter
    def quit(self, value):
        if value:
            self._quit = value
            self._pssofb.threads_shutdown()

    #  ----- log auxiliary methods -----

    def _update_log(self, msg):
        if 'ERR' in msg:
            _log.error(msg[4:])
        elif 'FATAL' in msg:
            _log.error(msg[6:])
        elif 'WARN' in msg:
            _log.warning(msg[5:])
        else:
            _log.info(msg)
        self.run_callbacks('Log-Mon', msg)

    # ----- auto save methods -----

    def read_autosave_file(self):
        """Read autosave file."""
        filename = self._const.autosave_fname
        config_name = None
        if _os.path.isfile(filename):
            with open(filename, 'r') as fil:
                config_name = fil.read().strip('\n')
        if config_name is not None:
            self._config_name = config_name
        else:
            _log.info(
                'No backup file was found, default value '
                'was used for si_idff configuration name.')
            self._config_name = self._get_default_configname()
            self.update_autosave_file()
            _log.info('First autosave file was created.')

    def update_autosave_file(self):
        """Update autosave file."""
        path = _os.path.split(self._const.autosave_fname)[0]
        _os.makedirs(path, exist_ok=True)
        with open(self._const.autosave_fname, 'w+') as fil:
            fil.write(self._config_name)

    def _get_default_configname(self):
        if self._const.idname.dev == 'EPU50':
            return 'epu50_ref'
        elif self._const.idname.dev == 'PAPU50':
            return 'papu50_ref'
        elif self._const.idname.dev == 'DELTA52':
            return 'delta52_ref'
        return ''

    # ----- update pvs methods -----

    def _do_sleep(self, time0, tplanned):
        ttook = _time.time() - time0
        tsleep = tplanned - ttook
        if tsleep > 0:
            _time.sleep(tsleep)
        else:
            _log.warning(
                'Feedforward step took more than planned... '
                '{0:.3f}/{1:.3f} s'.format(ttook, tplanned))

    def _do_update_polarization(self):
        new_pol, *_ = self._idff.get_polarization_state()
        if new_pol is not None and new_pol != self._polarization:
            self._polarization = new_pol
            self.run_callbacks('Polarization-Mon', new_pol)

    def _do_update_correctors(self):
        corrdevs = None
        if self._control_qs == self._const.DsblEnbl.Dsbl:
            corrdevs = self._idff.chdevs + self._idff.cvdevs
        try:
            # ret = self._idff.calculate_setpoints()
            # setpoints, polarization, *parameters = ret
            # pparameter_value, kparameter_value = parameters
            # print('pparameter: ', pparameter_value)
            # print('kparameter: ', kparameter_value)
            # print('polarization: ', polarization)
            # print('setpoints: ', setpoints)
            # print()
            if self._pssofb_isused:
                # calc setpoints
                ret = self._idff.calculate_setpoints()
                setpoints, *_ = ret

                # built curr_sp vector
                curr_sp = self._pssfob_get_current_setpoint(
                    setpoints, corrdevs)

                # apply curr_sp to pssofb
                self._pssofb.bsmp_sofb_current_set_update((curr_sp, ))
            else:
                # use PS IOCs (IDFF) to implement setpoints
                self._idff.implement_setpoints(corrdevs=corrdevs)

        except ValueError as err:
            self._update_log('ERR:'+str(err))

    def _do_ff(self):
        # updating loop
        while not self._quit:
            # updating interval
            tplanned = 1.0/self._loop_freq

            # initial time
            _t0 = _time.time()

            # check IDFF device connection
            if not self._idff.connected:
                self._do_sleep(_t0, tplanned)
                continue

            # update polarization state
            self._do_update_polarization()

            # return if loop is not closed
            if not self._loop_state:
                self._do_sleep(_t0, tplanned)
                continue

            # correctors setpoint implementation
            self._do_update_correctors()

            # sleep unused time or signal overtime to stdout
            self._do_sleep(_t0, tplanned)

    def _update_corr_status(self):
        """Update CorrStatus-Mon PV."""
        devs = self._idff.chdevs + self._idff.cvdevs
        if self._control_qs == self._const.DsblEnbl.Enbl:
            devs.extend(self._idff.qsdevs)

        status = 0
        if all(d.connected for d in devs):
            if any(d.pwrstate != d.PWRSTATE.On for d in devs):
                status = _updt_bit(status, 1, 1)
            if any(d.opmode != d.OPMODE_STS.SlowRef for d in devs):
                status = _updt_bit(status, 2, 1)
            if any(d.sofbmode != self._pssofb_isused for d in devs):
                status = _updt_bit(status, 3, 1)
        else:
            status = 0b111

        self.run_callbacks('CorrStatus-Mon', status)

    # ----- idff -----

    def _idff_prepare_corrs_state(self, pssofb_isused):
        """Configure PSSOFB mode state ."""
        if not self._idff.wait_for_connection():
            return False
        corrdevs = self._idff.chdevs + self._idff.cvdevs + self._idff.qsdevs
        for dev in corrdevs:
            if pssofb_isused:
                if not dev.cmd_sofbmode_enable(timeout=App.DEF_PS_TIMEOUT):
                    return False
            else:
                if not dev.cmd_sofbmode_disable(timeout=App.DEF_PS_TIMEOUT):
                    return False
        return True

    # ----- pssofb -----

    def _pssofb_init(self, idname):
        """Create PSSOFB connections to control correctors."""
        # bbbnames
        bbbnames = _PSNamesSOFB.get_bbbnames(idname)

        # create pssofb object
        pssofb = _PSConnSOFB(
            ethbridgeclnt_class=_EthBridgeClient,
            bbbnames=bbbnames,
            sofb_update_iocs=True,
            acc=idname)

        # build bsmpid -> psname dict
        bsmp_devs = dict()
        for devices in pssofb.bbb2devs.values():
            for devname, bsmpid in devices:
                bsmp_devs[bsmpid] = devname

        return pssofb, bsmp_devs

    def _pssfob_get_current_setpoint(self, setpoints, corrdevs):
        """Convert IDFF dict setpoints to PSSOFB list setpoints."""
        current_sp = _np.ones(len(setpoints)) * _np.nan
        devnames = [dev.devname for dev in corrdevs]
        for bsmpid, devname in self._bsmp_devs.items():
            if devname in devnames:
                current_sp[bsmpid - 1] = setpoints[devname]
        return current_sp



"""."""
import numpy as _np

from ..search import PSSearch as _PSSearch
from ..namesys import SiriusPVName as _SiriusPVName

from . import Device as _Device
from . import DeviceSet as _DeviceSet
from . psconv import StrengthConv as _StrengthConv


class PSNamesSOFB:
    """."""

    _sofb = dict()
    _sofb_factory = None

    @staticmethod
    def get_psnames_ch(acc):
        """Return horizontal corrector psnames of a given sector."""
        if PSNamesSOFB._sofb_factory is None:
            from ..sofb.csdev import SOFBFactory
            PSNamesSOFB._sofb_factory = SOFBFactory
        if acc not in PSNamesSOFB._sofb:
            PSNamesSOFB._sofb[acc] = PSNamesSOFB._sofb_factory.create(acc)
        return PSNamesSOFB._sofb[acc].ch_names

    @staticmethod
    def get_psnames_cv(acc):
        """Return vertical corrector psnames of a given sector."""
        if PSNamesSOFB._sofb_factory is None:
            from ..sofb.csdev import SOFBFactory
            PSNamesSOFB._sofb_factory = SOFBFactory
        if acc not in PSNamesSOFB._sofb:
            PSNamesSOFB._sofb[acc] = PSNamesSOFB._sofb_factory.create(acc)
        return PSNamesSOFB._sofb[acc].cv_names


class PSCorrSOFB(_Device):
    """SOFB corrector device.

    Group SOFB setpoints of all corrector power supplies belonging to
    the same BeagleBone.
    """

    _curr_sp = 'SOFBCurrent-SP'
    _curr_rb = 'SOFBCurrent-RB'
    _curr_refmon = 'SOFBCurrentRef-Mon'
    _curr_mon = 'SOFBCurrent-Mon'

    PROPERTIES_DEFAULT = (_curr_sp, _curr_rb, _curr_refmon, _curr_mon)

    def __init__(self, devname, auto_monitor_mon=False, props2init='all'):
        """."""
        self._devname_orig = _SiriusPVName(devname)
        self._sec = self._devname_orig.sec

        # check if device exists and is a SOFB corrector
        if devname not in PSNamesSOFB.get_psnames_ch(self._sec) and \
                devname not in PSNamesSOFB.get_psnames_cv(self._sec):
            raise NotImplementedError(devname)

        # get bbbname and linked bsmp devices
        self._bbbname = _PSSearch.conv_psname_2_bbbname(devname)
        self._bsmpdevs = _PSSearch.conv_bbbname_2_bsmps(self._bbbname)

        # call base class constructor
        devname = self._bsmpdevs[0][0]
        super().__init__(
            devname, props2init=props2init, auto_monitor_mon=auto_monitor_mon)

        # get sofb indices
        self._sofb_indices, self._idx_corr = self._get_sofb_indices()

    @property
    def devname(self):
        """Device name."""
        return self._devname_orig

    @property
    def devname_first_udc(self):
        """Device name of first UDC power supply."""
        return self._bsmpdevs[0][0]

    @property
    def bbbname(self):
        """BBB name."""
        return self._bbbname

    @property
    def bsmpdevs(self):
        """BSMP devices."""
        return self._bsmpdevs

    @property
    def sofb_indices(self):
        """Return indices of devices in SOFB current vector."""
        return self._sofb_indices

    @property
    def current(self):
        """Return current Ref-Mon in SOFB order."""
        # get refmon
        values = self[self._curr_refmon]

        # trim values
        values = _np.asarray(values)
        values = values[self._idx_corr]
        return values

    @current.setter
    def current(self, value):
        """Set current -SP in SOFB order."""
        # trim set value
        value = _np.asarray(value)
        idx_val = _np.where(~_np.isnan(value))[0]

        # combine refmon and  setpoint values
        values = self[self._curr_sp]
        idx_pv = self._idx_corr[idx_val]
        values[idx_pv] = value[idx_val]

        # set new values
        self[self._curr_sp] = values

    @property
    def current_sp(self):
        """Return current -SP in SOFB order."""
        # get setpoints
        values = self[self._curr_sp]

        # trim values
        values = _np.asarray(values)
        values = values[self._idx_corr]
        return values

    @property
    def current_rb(self):
        """Return current -RB in SOFB order."""
        # get rb
        values = self[self._curr_rb]

        # trim values
        values = _np.asarray(values)
        values = values[self._idx_corr]
        return values

    @property
    def current_mon(self):
        """Return current -Mon in SOFB order."""
        # get mon
        values = self[self._curr_mon]

        # trim values
        values = _np.asarray(values)
        values = values[self._idx_corr]
        return values

    # --- private methods ---

    def _get_sofb_indices(self):
        devids = sorted([dev[1] for dev in self._bsmpdevs])
        psnames = PSNamesSOFB.get_psnames_ch(self._sec) + \
            PSNamesSOFB.get_psnames_cv(self._sec)
        indices = list()
        idx_corr = list()
        for devname, dev_id in self._bsmpdevs:
            if devname in psnames:
                idx_id = devids.index(dev_id)
                idx_ps = psnames.index(devname)
                idx_corr.append(idx_id)
                indices.append(idx_ps)
        indices = _np.asarray(indices)
        idx_corr = _np.asarray(idx_corr)
        return indices, idx_corr


class PSApplySOFB(_DeviceSet):
    """SOFB corrector devices.

    Group SOFB setpoints of all corrector power supplies.
    """

    _dipole_propty = 'Ref-Mon'

    class DEVICES:
        """."""

        BO = 'BO'
        SI = 'SI'
        ALL = (BO, SI)

    def __init__(self, devname, auto_monitor_mon=False, dipoleoff=False):
        """."""
        # check if device exists
        if devname not in PSApplySOFB.DEVICES.ALL:
            raise NotImplementedError(devname)

        # get devices
        devices = PSApplySOFB._get_pscorrsofb_devices(
            devname, auto_monitor_mon)

        # strengthconv dictionaries
        self._pstype_2_index, self._pstype_2_sconv = \
            self._get_strenconv(devname, auto_monitor_mon)

        # add StrengthConv devices
        devices += self._pstype_2_sconv.values()

        # call base class constructor
        super().__init__(devices, devname=devname)

        # number of correctors
        self._nr_chs = len(self.psnames_ch)
        self._nr_cvs = len(self.psnames_cv)

        # dipole off: used to convert current<->kick with fixed 3GeV energy
        self._dipoleoff = dipoleoff

    @property
    def psnames_ch(self):
        """."""
        return PSNamesSOFB.get_psnames_ch(self.devname)

    @property
    def psnames_cv(self):
        """."""
        return PSNamesSOFB.get_psnames_cv(self.devname)

    @property
    def nr_correctors(self):
        """Return number of correctors."""
        return len(self.psnames_ch) + len(self.psnames_cv)

    @property
    def indices_ch(self):
        """Return selection indices for horizontal correctors."""
        return _np.arange(len(self.psnames_ch))

    @property
    def indices_cv(self):
        """Return selection indices for vertical correctors."""
        return _np.arange(len(self.psnames_ch), self.nr_correctors)

    @property
    def current(self):
        """Return CurrentRef-Mon vector in SOFB order."""
        currents = self._get_current('current')
        return currents

    @current.setter
    def current(self, value):
        """Set Current-SP vector in SOFB order."""
        value = _np.asarray(value)
        for corr in self.devices:
            if isinstance(corr, PSCorrSOFB):
                inds = corr.sofb_indices
                current = value[inds]
                corr.current = current

    @property
    def current_sp(self):
        """Return Current-SP vector in SOFB order."""
        currents = self._get_current('current_sp')
        return currents

    @property
    def current_rb(self):
        """Return Current-RB vector in SOFB order."""
        currents = self._get_current('current_rb')
        return currents

    @property
    def current_mon(self):
        """Return Current-Mon vector in SOFB order."""
        currents = self._get_current('current_mon')
        return currents

    @property
    def kick(self):
        """Return correctors Ref-Mon kicks in SOFB order."""
        current = self.current
        strength = self._conv_curr2stren(current)
        return strength

    @kick.setter
    def kick(self, value):
        """Set correctors -SP kicks in SOFB order."""
        current = self._conv_stren2curr(value)
        self.current = current

    @property
    def kick_sp(self):
        """Return correctors -SP kicks in SOFB order."""
        current = self.current_sp
        strength = self._conv_curr2stren(current)
        return strength

    @property
    def kick_rb(self):
        """Return correctors -RB kicks in SOFB order."""
        current = self.current_rb
        strength = self._conv_curr2stren(current)
        return strength

    @property
    def kick_mon(self):
        """Return correctors -Mon kicks in SOFB order."""
        current = self.current_mon
        strength = self._conv_curr2stren(current)
        return strength

    # --- private methods ---

    def _get_current(self, propty):
        values = _np.zeros(self._nr_chs + self._nr_cvs)
        for corr in self.devices:
            if isinstance(corr, PSCorrSOFB):
                inds = corr.sofb_indices
                vals = getattr(corr, propty)
                values[inds] = vals
        return values

    def _conv_curr2stren(self, current):
        strength = _np.zeros(len(current))
        for pstype, index in self._pstype_2_index.items():
            sconv = self._pstype_2_sconv[pstype]
            value = current[index]
            if self._dipoleoff:
                stren = sconv.conv_current_2_strength(
                    currents=value, strengths_dipole=3.0)
            else:
                stren = sconv.conv_current_2_strength(currents=value)
            strength[index] = stren
        return strength

    def _conv_stren2curr(self, strength):
        current = _np.full(len(strength), _np.nan, dtype=float)
        for pstype, index in self._pstype_2_index.items():
            sconv = self._pstype_2_sconv[pstype]
            value = strength[index]
            idcs = ~_np.isnan(value)
            if self._dipoleoff:
                curr = sconv.conv_strength_2_current(
                    strengths=value[idcs], strengths_dipole=3.0)
            else:
                curr = sconv.conv_strength_2_current(strengths=value[idcs])
            current[index[idcs]] = curr
        return current

    @staticmethod
    def _get_pscorrsofb_devices(devname, auto_monitor_mon):
        psnames = PSNamesSOFB.get_psnames_ch(devname) + \
            PSNamesSOFB.get_psnames_cv(devname)
        devices = dict()
        all_devices = list()
        for psname in psnames:
            if psname in all_devices:
                continue
            sofb_corr = PSCorrSOFB(psname, auto_monitor_mon)
            all_devices += [dev[0] for dev in sofb_corr.bsmpdevs]
            devname = sofb_corr.devname_first_udc
            if devname not in devices:
                devices[devname] = sofb_corr
        return list(devices.values())

    def _get_strenconv(self, devname, auto_monitor_mon):
        # 1. create pstype to StrengthConv dictionary.
        # 2. create pstype to corrector index dictionnary.
        pstype_2_index = dict()
        pstype_2_sconv = dict()
        psnames = PSNamesSOFB.get_psnames_ch(devname) + \
            PSNamesSOFB.get_psnames_cv(devname)
        for i, psname in enumerate(psnames):
            pstype = _PSSearch.conv_psname_2_pstype(psname)
            if pstype not in pstype_2_index:
                pstype_2_index[pstype] = []
            pstype_2_index[pstype].append(i)
            if pstype not in pstype_2_sconv:
                sconv = _StrengthConv(
                    psname, PSApplySOFB._dipole_propty,
                    auto_monitor_mon=auto_monitor_mon)
                pstype_2_sconv[pstype] = sconv

        # convert index to numpy array
        for pstype in pstype_2_index:
            pstype_2_index[pstype] = _np.asarray(pstype_2_index[pstype])

        return pstype_2_index, pstype_2_sconv



"""Power Supply BSMP Commands.

Documentation:

https://wiki-sirius.lnls.br/mediawiki/index.php/Machine:Power_Supplies
"""

import time as _time
import numpy as _np

from ...bsmp import constants as _const_bsmp
from ...bsmp import (
    BSMP as _BSMP,
    IOInterface as _IOInterface,
)

from . import constants as _const_psbsmp
from . import entities as _etity_psbsmp
from . entities import EntitiesPS as _EntitiesPS

# version of the BSMP implementation of power supplies that is compatible
# with the current implemenation of this module.


class PSBSMP(_BSMP):
    """Power supply BSMP.

    This class derives from basic BSMP class.
    It overrides methods with power supply specific parameters.
    """

    CONST_BSMP = _const_bsmp
    CONST = _const_psbsmp.ConstPSBSMP

    _timeout_read_variable = 100  # [ms]
    _timeout_execute_function = 100  # [ms]
    _timeout_remove_vars_groups = 100  # [ms]
    _timeout_create_vars_groups = 100  # [ms]
    _timeout_read_group_of_variables = 100  # [ms]
    _timeout_request_curve_block = 100  # [ms]
    _timeout_curve_block = 100  # [ms]
    _timeout_query_list_of_group_of_variables = 100  # [ms]
    _timeout_query_group_of_variables = 100  # [ms]

    _sleep_turn_onoff = 0.050  # [s]
    _sleep_reset_udc = 1.000  # [s]
    _sleep_disable_scope = 0.5  # [s]
    _sleep_select_op_mode = 0.030  # [s]

    # --- BSMP PS curves ---
    CURVE_ID_WFMREF_DATA0 = 0
    CURVE_ID_WFMREF_DATA1 = 1
    CURVE_ID_SCOPE = 2

    _WFMREF_POINTERS_VAR_IDS = {
        CURVE_ID_WFMREF_DATA0: (
            CONST.V_WFMREF0_START,
            CONST.V_WFMREF0_END,
            CONST.V_WFMREF0_IDX),
        CURVE_ID_WFMREF_DATA1: (
            CONST.V_WFMREF1_START,
            CONST.V_WFMREF1_END,
            CONST.V_WFMREF1_IDX),
    }

    def __init__(self, slave_address: int, entities: _EntitiesPS, pru: _IOInterface):
        """Init BSMP."""
        self.pru: _IOInterface = pru
        self.entities: _EntitiesPS
        super().__init__(self.pru, slave_address, entities)
        self._wfmref_check_entities_consistency()

    @staticmethod
    def parse_firmware_version(version) -> str:
        """Process firmware version from BSMP device."""
        version = version[:version.index(b'\x00')]
        version = ''.join([chr(ord(v)) for v in version])
        return version

    def reset_groups_of_variables(self, varids_groups):
        """Reset groups of variables."""

        # remove previous variables groups
        ack, data = self.remove_all_groups_of_variables(
            timeout=self._timeout_remove_vars_groups)
        if ack != PSBSMP.CONST_BSMP.ACK_OK:
            return ack, data

        # create variables groups
        for var_ids in varids_groups:
            ack, data = self.create_group_of_variables(
                var_ids, timeout=self._timeout_create_vars_groups)
            if ack != PSBSMP.CONST_BSMP.ACK_OK:
                return ack, data

        return ack, data

    # --- bsmp overriden methods ---

    def execute_function(
            self, func_id, input_val=None,
            timeout=_timeout_execute_function,
            read_flag=True, print_error=True):
        """."""
        if func_id in (PSBSMP.CONST.F_SYNC_PULSE,
                       PSBSMP.CONST.F_RESET_UDC):
            # NOTE: this is necessary while PS BSMP spec is
            # not updated!
            read_flag = False

        response = super().execute_function(
            func_id=func_id, input_val=input_val,
            timeout=timeout, read_flag=read_flag)

        # introduces necessary sleeps
        # TODO: check with ELP if these numbers can be optimized further!
        if func_id == PSBSMP.CONST.F_RESET_UDC:
            _time.sleep(self._sleep_reset_udc)
        elif func_id == PSBSMP.CONST.F_DISABLE_SCOPE:
            # NOTE: sleep is implemented in UDC class,
            # for optimization purpose!
            # _time.sleep(self._sleep_disable_scope)
            pass
        elif func_id == PSBSMP.CONST.F_SELECT_OP_MODE:
            # _time.sleep(self._sleep_select_op_mode)
            pass
        elif func_id in (PSBSMP.CONST.F_TURN_ON,
                         PSBSMP.CONST.F_TURN_OFF,
                         PSBSMP.CONST.F_OPEN_LOOP,
                         PSBSMP.CONST.F_CLOSE_LOOP):
            _time.sleep(self._sleep_turn_onoff)

        if read_flag:
            return response
        else:
            # TODO: This should be temporary. It is used for F_RESET_UDC.
            # Firmware should change so that F_RESET_UDC returns ack.
            return self.CONST_BSMP.ACK_OK, None

    def read_group_of_variables(
            self, group_id, timeout=_timeout_read_group_of_variables):
        """."""
        response = super().read_group_of_variables(
            group_id=group_id, timeout=timeout)

        return response

    def query_list_of_group_of_variables(
            self, timeout=_timeout_query_list_of_group_of_variables):
        """."""
        response = super().query_list_of_group_of_variables(
            timeout=timeout)

        return response

    def query_group_of_variables(
            self, group_id, timeout=_timeout_query_group_of_variables):
        """."""
        response = super().query_group_of_variables(
            group_id=group_id, timeout=timeout)

        return response

    # --- pwrsupply functions ---

    def ps_function_set_slowref_fbp(self, setpoints):
        """Write FBP 4-valued setpoints."""
        self.execute_function(
            PSBSMP.CONST.F_SET_SLOWREF_FBP, setpoints)

    def ps_function_set_slowref_fbp_readback_ref(self, setpoints):
        """Write FBP 4-valued setpoints and get reference values."""
        ack, data = self.execute_function(
            PSBSMP.CONST.F_SET_SLOWREF_FBP_READBACK_REF, setpoints)
        return ack, data

    # --- pwrsupply parameters ---
    def _parameter_read_ps_name(self, count):
        # PS_NAME parameter
        value = _np.zeros(count)
        for index in range(count):
            value[index] = self._read_parm(PSBSMP.CONST.P_PS_NAME, index)
            if value[index] == 0.0:
                # TODO: this end-of-string convention is
                # yet to be implemented!
                value = value[:index]
                break
        return ''.join([chr(int(v)) for v in value]).strip()

    def parameter_read(self, eid, index=None):
        """Return power supply parameter."""

        parameter = self.entities.parameters[eid]
        count = parameter['count']

        if eid == PSBSMP.CONST.P_PS_NAME:
            # PS_NAME parameter
            return self._parameter_read_ps_name(count=count)

        # other parameters
        if index is None:
            value = _np.zeros(count)
            for idx in range(count):
                val = self._read_parm(eid, idx)
                if val is None:
                    return None
                value[idx] = val
            if count == 1:
                value = value[0]
        else:
            value = self._read_parm(eid, index)
            if value is None:
                return None
        return value

    def _parameter_write_ps_name(self, count, val):
        value = _np.zeros(count)
        value[:len(val)] = [float(ord(c)) for c in val]
        for index, value_datum in enumerate(value):
            self.execute_function(
                func_id=PSBSMP.CONST.F_SET_PARAM,
                input_val=(PSBSMP.CONST.P_PS_NAME, index, value_datum),
            )

    def parameter_write(self, eid, value, index=0) -> None:
        """Set power supply parameter."""
        parameter = self.entities.parameters[eid]
        count = parameter['count']
        val = value[:count]
        if eid == PSBSMP.CONST.P_PS_NAME:
            # PS_NAME parameter
            self._parameter_write_ps_name(count=count, val=val)
        else:
            self.execute_function(
                func_id=PSBSMP.CONST.F_SET_PARAM,
                input_val=(eid, index, value))

    # --- wfmref and scope methods ---

    @property
    def wfmref_select(self):
        """Return wfmref curve ID currently in use by DSP."""
        _, curve_id = self.read_variable(
            var_id=PSBSMP.CONST.V_WFMREF_SELECTED,
            timeout=self._timeout_read_variable)
        return curve_id

    @wfmref_select.setter
    def wfmref_select(self, curve_id):
        """Select wfmref curve ID to be used by DSP."""
        ack, data = self.execute_function(
            func_id=PSBSMP.CONST.F_SELECT_WFMREF,
            input_val=curve_id,
            timeout=self._timeout_execute_function)
        return ack, data

    def wfmref_size(self, curve_id=None):
        """Return wfmref curve size in t_float units.

            This is the waveform size as last registered by the
        ARM controller.
        """
        # calculate wfmref size from buffer pointer values used by
        # ARM controller
        i_beg, i_end, _ = self._wfmref_get_pointers_ids_of_selected(curve_id)
        values = self._bsmp_get_variable_values(i_beg, i_end)
        return PSBSMP.curve_index_calc(values[0], values[1])

    @property
    def wfmref_index(self):
        """Return wfmref curve index.

            This index refers to the current waveform in use by the
        DSP controller.
        """
        # calculate wfmref index from buffer pointer values used by
        # ARM controller
        i_beg, _, i_idx = self._wfmref_get_pointers_ids_of_selected()
        values = self._bsmp_get_variable_values(i_beg, i_idx)
        return PSBSMP.curve_index_calc(values[0], values[1])

    @property
    def wfmref_maxsize(self):
        """Return max size of bsmp curves 0 and 1."""
        # curve with ids 0 and 1 should have same sizes.
        maxsize = self.curve_maxsize(curve_id=0)
        return maxsize

    @ property
    def wfmref_pointer_values(self):
        """Return pointer values of currently selected wfmref curve."""
        pointer_ids = self._wfmref_get_pointers_ids_of_selected()
        pointer_values = self._bsmp_get_variable_values(*pointer_ids)
        return pointer_values

    def wfmref_read(self):
        """Return data of wfmref curve currently in use by DSP."""
        # get curve ID
        curve_id = self.wfmref_select

        # get curve data
        curve = self._curve_bsmp_read(curve_id=curve_id)
        return curve

    def wfmref_write(self, curve):
        """Write a new wfmref curve to the available buffer not in use."""
        # get id of writable wfmref curve
        curve_id = self._wfmref_bsmp_select_writable_curve_id()

        # write curve
        curve = self.curve_write(curve_id, curve, read_curve=True)

        # execute selection of WfmRef to be used
        self.wfmref_select = curve_id

        return curve

    def scope_read(self):
        """Read scope."""
        curve_id = PSBSMP.CURVE_ID_SCOPE
        curve = self._curve_bsmp_read(curve_id=curve_id)
        return curve

    def scope_enable(self):
        """Enable scope update."""
        ack, data = self.execute_function(
            func_id=PSBSMP.CONST.F_ENABLE_SCOPE,
            timeout=self._timeout_execute_function)
        return ack, data

    def scope_disable(self):
        """Disable scope update."""
        ack, data = self.execute_function(
            func_id=PSBSMP.CONST.F_DISABLE_SCOPE,
            timeout=self._timeout_execute_function)
        return ack, data

    # --- curve methods ---

    def curve_read(self, curve_id):
        """Return curve."""
        curve = self._curve_bsmp_read(curve_id=curve_id)
        return curve

    def curve_write(self, curve_id, curve, read_curve=False):
        """Write curve."""
        # check curve size
        self._curve_check_size(curve_id=curve_id, curve_size=len(curve))

        # send curve blocks
        self._curve_bsmp_write(curve_id, curve)

        if read_curve:
            # get curve just written
            curve = self.curve_read(curve_id)
            return _np.array(curve)
        else:
            return curve

    def curve_maxsize(self, curve_id):
        """Return max size if t_float units according to BSMP spec."""
        curve_entity = self.entities.curves[curve_id]
        maxsize = curve_entity.max_size_t_float
        return maxsize

    @staticmethod
    def curve_index_calc(beg, end):
        """Calculate curve index."""
        return 1 + (end - beg) // 2

    # --- private methods ---

    def _wfmref_get_pointers_ids_of_selected(self, curve_id=None):
        if curve_id is None:
            curve_id = self.wfmref_select
        return PSBSMP._WFMREF_POINTERS_VAR_IDS[curve_id]

    def _wfmref_check_entities_consistency(self):
        # check consistency of curves with ids 0 and 1
        curves = self.entities.curves
        if 0 in curves and 1 in curves:
            curve0, curve1 = curves[0], curves[1]
            if False in (curve0.waccess, curve1.waccess) or \
               curve0.size != curve1.size or \
               curve0.nblocks != curve1.nblocks:
                raise ValueError('Inconsistent curves!')

    def _wfmref_bsmp_select_writable_curve_id(self):

        # get current curve id
        _, curve_id = self.read_variable(
            var_id=PSBSMP.CONST.V_WFMREF_SELECTED,
            timeout=self._timeout_read_variable)

        # select the other buffer and send curve blocks
        if curve_id == PSBSMP.CURVE_ID_WFMREF_DATA1:
            curve_id = PSBSMP.CURVE_ID_WFMREF_DATA0
        else:
            curve_id = PSBSMP.CURVE_ID_WFMREF_DATA1

        return curve_id

    def _curve_bsmp_read(self, curve_id):
        # select minimum curve size between spec and firmware.
        if curve_id == PSBSMP.CURVE_ID_SCOPE:
            # for scope, we have to assume the size as the same as selected
            # buffer since there are no bsmp variables indicating its size.
            wfmref_size = self.wfmref_size()
        else:
            wfmref_size = self.wfmref_size(curve_id)
        wfmref_size_min = self._curve_get_implementable_size(
            curve_id, wfmref_size)

        # create initial output data
        curve = _np.zeros(wfmref_size_min)

        # read curve blocks
        curve_entity = self.entities.curves[curve_id]
        indices = curve_entity.get_indices(wfmref_size_min)
        # print('reading - curve id: ', curve_id)
        # print('reading - indices: ', indices)
        for block, idx in enumerate(indices):
            # print('psbsmp.curve_read-0 ', curve_id, block)
            ack, data = self.request_curve_block(
                curve_id=curve_id,
                block=block,
                timeout=self._timeout_request_curve_block,
                print_error=False)
            # print(sum(data))
            # print((hex(ack), sum(data)))
            # print('psbsmp.curve_read-1')
            if ack != self.CONST_BSMP.ACK_OK:
                # print('psbsmp.curve_read-2')
                if curve_id == PSBSMP.CURVE_ID_SCOPE and \
                   ack == self.CONST_BSMP.ACK_RESOURCE_BUSY:
                    # print('sit1, add:{}, curve_id:{}, block:{}'.format(
                    #     add, curve_id, block))
                    # This is the expected behaviour when DSP is writting to
                    # buffer sample
                    return None
                # anomalous response!
                PSBSMP.anomalous_response(
                    self.CONST_BSMP.CMD_REQUEST_CURVE_BLOCK, ack,
                    curve_len=len(curve),
                    curve_id=curve_id,
                    block=block,
                    index=idx)
                # print('sit2, add:{}, curve_id:{}, block:{}'.format(
                #     add, curve_id, block))
                return None
            else:
                # print('sit3, add:{}, curve_id:{}, block:{}'.format(
                #     add, curve_id, block))
                idx_w = slice(idx[0], idx[1])
                idx_r = slice(0, idx[1] - idx[0])
                curve[idx_w] = data[idx_r]
        return curve

    def _curve_bsmp_write(self, curve_id, curve):
        # select minimum curve size between spec and firmware.
        wfmref_size_min = self._curve_get_implementable_size(
            curve_id, len(curve))

        # send curve blocks
        curve_entity = self.entities.curves[curve_id]
        indices = curve_entity.get_indices(wfmref_size_min)
        # print('writing - curve id: ', curve_id)
        # print('writing - indices: ', indices)

        # time0 = _time.time()
        for block, idx in enumerate(indices):
            data = curve[idx[0]:idx[1]]
            # print(sum(data))
            ack, data = self.curve_block(
                curve_id=curve_id,
                block=block,
                value=data,
                timeout=self._timeout_curve_block)
            # print((hex(ack), data))
            if ack != self.CONST_BSMP.ACK_OK:
                print(('BSMP response not OK in '
                       '_curve_bsmp_write: ack = 0x{:02X}!').format(ack))
                PSBSMP.anomalous_response(
                    self.CONST_BSMP.CMD_CURVE_BLOCK, ack,
                    curve_len=len(curve),
                    curve_id=curve_id,
                    block=block,
                    index=idx)
        # time1 = _time.time()
        # print('blocks write: {:.4f} ms'.format(1000*(time1 - time0)))

    def _bsmp_get_variable_values(self, *var_ids):
        values = [None] * len(var_ids)
        for i, var_id in enumerate(var_ids):
            _, values[i] = self.read_variable(
                var_id=var_id, timeout=self._timeout_read_variable)
        return values

    def _curve_get_implementable_size(self, curve_id, curve_size):
        curve_entity = self.entities.curves[curve_id]
        curve_size_entity = curve_entity.max_size_t_float
        curve_size_min = min(curve_size_entity, curve_size)
        return curve_size_min

    def _curve_check_size(self, curve_id, curve_size):
        if curve_size > self.curve_maxsize(curve_id=curve_id):
            raise IndexError('Curve exceeds maximum size according to spec!')

    def _read_parm(self, eid, idx):
        ack, val = self.execute_function(
            func_id=PSBSMP.CONST.F_GET_PARAM,
            input_val=(eid, idx))
        if ack == PSBSMP.CONST_BSMP.CMD_FUNCTION_ERROR:
            # TODO: when a uninitialized parameter is requested
            # the UDC responds with a function error.
            # this is a temporary bug bypass while firmware is not
            # updated!
            val = float('NaN')
        elif ack != PSBSMP.CONST_BSMP.ACK_OK:
            return None
        return val


# --- DCDC ---


class FBP(PSBSMP):
    """BSMP with EntitiesFBP."""

    SOFB_PRINT_COMM_ERRORS = True

    IS_DCLINK = False
    CONST = _const_psbsmp.ConstFBP
    _ACK_OK = _np.zeros(_const_psbsmp.UDC_MAX_NR_DEV, dtype=int)

    def __init__(self, slave_address: int, pru: _IOInterface):
        """Init BSMP."""
        PSBSMP.__init__(
            self, slave_address, _etity_psbsmp.EntitiesFBP(), pru=pru)

        # SOFB attributes
        self._sofb_ps_setpoint = None
        self._sofb_ps_reference = None
        self._sofb_ps_iload = None
        self._sofb_ps_readback_ref = None
        self._sofb_ps_func_return = None

    # --- SOFB methods ---

    @property
    def sofb_ps_setpoint(self):
        """."""
        return self._sofb_ps_setpoint

    @property
    def sofb_ps_reference(self):
        """."""
        return self._sofb_ps_reference

    @property
    def sofb_ps_iload(self):
        """."""
        return self._sofb_ps_iload

    @property
    def sofb_ps_readback_ref(self):
        """Return mirror powersupply currents read after last setpoint."""
        return self._sofb_ps_readback_ref

    @property
    def sofb_ps_func_return(self):
        """Return ack code of last function execution."""
        return self._sofb_ps_func_return

    def sofb_ps_setpoint_set(self, value):
        """."""
        if self._sofb_ps_setpoint is None:
            self._sofb_ps_setpoint = _np.zeros(
                _const_psbsmp.UDC_MAX_NR_DEV, dtype=float)

        # NOTE: NaNs in current vector are a way to avoid communication with
        #   udcs. In case only a few inputs are NaN they will be replaced by
        #   old values or zero. If all of them are NaN, them no communication
        #   is performed.
        nan = _np.isnan(value)
        self._sofb_ps_setpoint[~nan] = value[~nan]

        # NOTE: In case all values are NaN, no communication is made:
        if nan.all():
            ack = self.CONST_BSMP.ACK_OK
            func_resp = self._sofb_ps_readback_ref
        else:
            ack, func_resp = self.ps_function_set_slowref_fbp_readback_ref(
                self._sofb_ps_setpoint)

        if ack == self.CONST_BSMP.ACK_OK:
            self._sofb_ps_readback_ref = func_resp
            self._sofb_ps_func_return = FBP._ACK_OK
        else:
            self._sofb_func_error(ack, func_resp, 'sofb_ps_setpoint_set')

    def sofb_update(self):
        """."""
        data = self._sofb_read_group_of_variables()
        (self._sofb_ps_setpoint,
         self._sofb_ps_reference,
         self._sofb_ps_iload) = data

    def _sofb_read_group_of_variables(self):
        group_id = self.CONST.G_SOFB
        ack, func_resp = self.read_group_of_variables(
            group_id=group_id)
        if ack == self.CONST_BSMP.ACK_OK:
            self._sofb_ps_func_return = FBP._ACK_OK
            setpoints, references, iload = \
                _np.array(func_resp).reshape((3, -1))
        else:
            setpoints, references, iload = None, None, None
            self._sofb_func_error(
                ack, func_resp, '_sofb_read_group_of_variables')

        return setpoints, references, iload

    def _sofb_func_error(self, ack, func_resp, methodname):
        func_resp = ord(func_resp)
        self._sofb_ps_func_return = FBP._ACK_OK + func_resp
        if self.SOFB_PRINT_COMM_ERRORS:
            sfmt = 'FBP: Anomalous response ' + methodname + \
                ': ack:0x{:02X}, func_resp:0x{:02X}'
            print(sfmt.format(ack, func_resp))


class FAC_DCDC(PSBSMP):
    """BSMP with EntitiesFAC_DCDC."""

    IS_DCLINK = False
    CONST = _const_psbsmp.ConstFAC_DCDC

    def __init__(self, slave_address: int, pru: _IOInterface):
        """Init BSMP."""
        PSBSMP.__init__(
            self, slave_address, _etity_psbsmp.EntitiesFAC_DCDC(), pru=pru)


class FAC_2P4S_DCDC(PSBSMP):
    """BSMP with EntitiesFAC_2P4S_DCDC."""

    IS_DCLINK = False
    CONST = _const_psbsmp.ConstFAC_2P4S_DCDC

    def __init__(self, slave_address: int, pru: _IOInterface):
        """Init BSMP."""
        PSBSMP.__init__(
            self, slave_address, _etity_psbsmp.EntitiesFAC_2P4S_DCDC(),
            pru=pru)


class FAC_2S_DCDC(PSBSMP):
    """BSMP with EntitiesFAC_2S_DCDC."""

    IS_DCLINK = False
    CONST = _const_psbsmp.ConstFAC_2S_DCDC

    def __init__(self, slave_address: int, pru: _IOInterface):
        """Init BSMP."""
        PSBSMP.__init__(
            self, slave_address, _etity_psbsmp.EntitiesFAC_2S_DCDC(), pru=pru)


class FAP(PSBSMP):
    """BSMP with EntitiesFAP."""

    IS_DCLINK = False
    CONST = _const_psbsmp.ConstFAP

    def __init__(self, slave_address: int, pru: _IOInterface):
        """Init BSMP."""
        PSBSMP.__init__(
            self, slave_address, _etity_psbsmp.EntitiesFAP(), pru=pru)


class FAP_4P(PSBSMP):
    """BSMP with EntitiesFAP_4P."""

    IS_DCLINK = False
    CONST = _const_psbsmp.ConstFAP_4P

    def __init__(self, slave_address: int, pru: _IOInterface):
        """Init BSMP."""
        PSBSMP.__init__(
            self, slave_address, _etity_psbsmp.EntitiesFAP_4P(), pru=pru)


class FAP_2P2S(PSBSMP):
    """BSMP with EntitiesFAP_2P2S."""

    IS_DCLINK = False
    CONST = _const_psbsmp.ConstFAP_2P2S

    def __init__(self, slave_address: int, pru: _IOInterface):
        """Init BSMP."""
        PSBSMP.__init__(
            self, slave_address, _etity_psbsmp.EntitiesFAP_2P2S(), pru=pru)


# --- ACDC ---


class FBP_DCLink(PSBSMP):
    """BSMP with EntitiesFBP_DCLink."""

    IS_DCLINK = True
    CONST = _const_psbsmp.ConstFBP_DCLink

    def __init__(self, slave_address: int, pru: _IOInterface):
        """Init BSMP."""
        PSBSMP.__init__(
            self, slave_address, _etity_psbsmp.EntitiesFBP_DCLink(), pru=pru)


class FAC_2P4S_ACDC(PSBSMP):
    """BSMP with EntitiesFAC_2P4S_ACDC."""

    IS_DCLINK = True
    CONST = _const_psbsmp.ConstFAC_2P4S_ACDC

    def __init__(self, slave_address: int, pru: _IOInterface):
        """Init BSMP."""
        PSBSMP.__init__(
            self, slave_address, _etity_psbsmp.EntitiesFAC_2P4S_ACDC(),
            pru=pru)


class FAC_2S_ACDC(PSBSMP):
    """BSMP with EntitiesFAC_2S_ACDC."""

    IS_DCLINK = True
    CONST = _const_psbsmp.ConstFAC_2S_ACDC

    def __init__(self, slave_address: int, pru: _IOInterface):
        """Init BSMP."""
        PSBSMP.__init__(
            self, slave_address, _etity_psbsmp.EntitiesFAC_2S_ACDC(), pru=pru)



"""Accelerator lattice and optics constants."""


class Accelerator:
    """Accelerator."""


class Ring(Accelerator):
    """Ring."""


class SI(Ring):
    """SI."""

    length = 518.396  # [m]
    circumference = length
    harmonic_number = 864


class BO(Ring):
    """BO."""

    length = 496.396  # [m]
    circumference = length
    harmonic_number = 828
